@@ -1,8 +1,9 @@
 typedef struct mthread {
 	PerlInterpreter* interp;
 	perl_mutex lock;
-	message_queue queue;
+	message_queue* queue;
 	UV id;
+	bool alive;
 #ifndef WIN32
 	sigset_t initial_sigmask;   /* Thread wakes up with signals blocked */
 #endif
