@@ -9,6 +9,14 @@
  * Message queues
  */
 
+typedef struct {
+	message_queue parent;
+	perl_mutex mutex;
+	perl_cond condvar;
+	message* front;
+	message* back;
+} message_queue_impl ;
+
 static void node_unshift(message** position, message* new_node) {
 	new_node->next = *position;
 	*position = new_node;
@@ -39,13 +47,8 @@ static void S_node_destroy(pTHX_ message** current) {
 }
 #define node_destroy(current) S_node_destroy(aTHX_ current)
 
-void queue_init(message_queue* queue) {
-	Zero(queue, 1, message_queue);
-	MUTEX_INIT(&queue->mutex);
-	COND_INIT(&queue->condvar);
-}
-
-void S_queue_enqueue(pTHX_ message_queue* queue, const message* message_, perl_mutex* external_lock) {
+static void S_queue_enqueue(pTHX_ message_queue* _queue, const message* message_, perl_mutex* external_lock) {
+	message_queue_impl* queue = (message_queue_impl*) _queue;
 	message* new_entry;
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
@@ -59,7 +62,8 @@ void S_queue_enqueue(pTHX_ message_queue* queue, const message* message_, perl_m
 	MUTEX_UNLOCK(&queue->mutex);
 }
 
-static const message* queue_shift(message_queue* queue) {
+static const message* queue_shift(message_queue_impl* _queue) {
+	message_queue_impl* queue = (message_queue_impl*) _queue;
 	const message* ret = node_shift(&queue->front);
 
 	if (queue->front == NULL)
@@ -67,7 +71,8 @@ static const message* queue_shift(message_queue* queue) {
 	return ret;
 }
 
-const message* S_queue_dequeue(pTHX_ message_queue* queue, perl_mutex* external_lock) {
+static const message* S_queue_dequeue(pTHX_ message_queue* _queue, perl_mutex* external_lock) {
+	message_queue_impl* queue = (message_queue_impl*) _queue;
 	const message* ret;
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
@@ -82,7 +87,8 @@ const message* S_queue_dequeue(pTHX_ message_queue* queue, perl_mutex* external_
 	return ret;
 }
 
-const message* S_queue_dequeue_nb(pTHX_ message_queue* queue, perl_mutex* external_lock) {
+static const message* S_queue_dequeue_nb(pTHX_ message_queue* _queue, perl_mutex* external_lock) {
+	message_queue_impl* queue = (message_queue_impl*) _queue;
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
 		MUTEX_UNLOCK(external_lock);
@@ -99,10 +105,28 @@ const message* S_queue_dequeue_nb(pTHX_ message_queue* queue, perl_mutex* extern
 	}
 }
 
-void S_queue_destroy(pTHX_ message_queue* queue) {
+static void S_queue_destroy(pTHX_ message_queue* _queue) {
+	message_queue_impl* queue = (message_queue_impl*) _queue;
 	MUTEX_LOCK(&queue->mutex);
 	node_destroy(&queue->front);
 	COND_DESTROY(&queue->condvar);
 	MUTEX_UNLOCK(&queue->mutex);
 	MUTEX_DESTROY(&queue->mutex);
+	PerlMemShared_free(queue);
+}
+
+const message_queue_vtable simple_table = {
+	S_queue_enqueue,
+	S_queue_dequeue,
+	S_queue_dequeue_nb,
+	S_queue_destroy
+};
+
+message_queue* S_queue_simple_alloc(pTHX) {
+	message_queue_impl* ret = PerlMemShared_calloc(1, sizeof(message_queue_impl));
+	Zero(ret, 1, message_queue);
+	ret->parent.table = &simple_table;
+	MUTEX_INIT(&ret->mutex);
+	COND_INIT(&ret->condvar);
+	return (message_queue*)ret;
 }
