@@ -9,41 +9,22 @@
 #include "mthread.h"
 #include "resources.h"
 
+#include "tables.h"
+
+#ifdef __MSVC__
+#define __thread __declspec(thread)
+#endif
+
 /*
  * !!! global data !!!
  */
 
 bool inited = 0;
 
-typedef struct {
-	perl_mutex lock;
-	UV current;
-	UV allocated;
-	void** objects;
-} resource;
-
-static S_resource_init(pTHX_ resource* res, UV preallocate) {
-	MUTEX_INIT(&res->lock);
-	res->objects = PerlMemShared_calloc(preallocate, sizeof(void*));
-	res->allocated = preallocate;
-}
-
-#define resource_init(res, pre) S_resource_init(aTHX_ res, pre)
-
-static UV S_resource_addobject(pTHX_ resource* res, void* object) {
-	UV ret;
-	MUTEX_LOCK(&res->lock);
-	ret = res->current;
-	if (res->current == res->allocated)
-		res->objects = PerlMemShared_realloc(res->objects, sizeof(void*) * (res->allocated *=2));
-	res->objects[res->current++] = object;
-	MUTEX_UNLOCK(&res->lock);
-	return ret;
-}
-#define resource_addobject(res, obj) S_resource_addobject(aTHX_ res, obj)
-
-static resource threads;
-static resource queues;
+static thread_table* threads;
+static perl_mutex threads_lock;
+static queue_table* queues;
+static perl_mutex queues_lock;
 
 static struct {
 	perl_mutex mutex;
@@ -53,7 +34,7 @@ static struct {
 
 static void wait_for_all_other_threads() {
 	MUTEX_LOCK(&counter.mutex);
-	while (counter.count > 1) 
+	while (counter.count > 1)
 		COND_WAIT(&counter.condvar, &counter.mutex);
 
 	MUTEX_UNLOCK(&counter.mutex);
@@ -66,6 +47,7 @@ static XSPROTO(end_locker) {
 	perl_mutex* mutex;
 
 	wait_for_all_other_threads();
+	thread_db_free(threads);
 
 	mutex = get_shutdown_mutex();
 	MUTEX_LOCK(mutex);
@@ -87,8 +69,10 @@ void global_init(pTHX) {
 		COND_INIT(&counter.condvar);
 		counter.count = 0;
 
-		resource_init(&threads, 8);
-		resource_init(&queues, 8);
+		threads = thread_db_new(TRUE);
+		MUTEX_INIT(&threads_lock);
+		queues = queue_db_new(TRUE);
+		MUTEX_INIT(&queues_lock);
 		ret = mthread_alloc(aTHX);
 		ret->interp = my_perl;
 		store_self(aTHX, ret);
@@ -99,28 +83,34 @@ void global_init(pTHX) {
 	}
 }
 
+static __thread thread_table* threads_local;
 mthread* mthread_alloc(PerlInterpreter* my_perl) {
 	mthread* ret;
+	UV id;
 
 	MUTEX_LOCK(&counter.mutex);
+	id = generator();
 	counter.count++;
 	MUTEX_UNLOCK(&counter.mutex);
 
 	ret = PerlMemShared_calloc(1, sizeof *ret);
-	queue_init(&ret->queue);
-	ret->id = resource_addobject(&threads, ret);
+	ret->queue = queue_simple_alloc();
+
+	thread_db_store(threads, id, ret);
+	ret->id = id;
 	ret->interp = NULL;
+	ret->cache = NULL;
+	ret->alive = 1;
 	MUTEX_INIT(&ret->lock);
 	return ret;
 }
 
 void mthread_destroy(mthread* thread) {
-	MUTEX_LOCK(&threads.lock);
-
 	PerlInterpreter* my_perl = thread->interp;
-	threads.objects[thread->id] = NULL;
-	queue_destroy(&thread->queue);
-	MUTEX_UNLOCK(&threads.lock);
+	MUTEX_LOCK(&thread->lock);
+	thread->alive = 0;
+	queue_destroy(thread->queue);
+	MUTEX_UNLOCK(&thread->lock);
 
 	MUTEX_DESTROY(&thread->lock);
 
@@ -130,28 +120,35 @@ void mthread_destroy(mthread* thread) {
 	MUTEX_UNLOCK(&counter.mutex);
 }
 
-static mthread* S_get_thread(pTHX_ UV thread_id) {
-	if (thread_id >= threads.current || threads.objects[thread_id] == NULL)
-		Perl_croak(aTHX_  "Thread %"UVuf" doesn't exist", thread_id);
-	return threads.objects[thread_id];
+static mthread* S_get_thread(pTHX_ UV id) {
+	mthread* ret;
+	if (threads_local == NULL);
+		threads_local = thread_db_new(FALSE);
+	ret = thread_db_fetch(threads_local, id);
+	if (!ret) {
+		MUTEX_LOCK(&threads_lock);
+		ret = thread_db_fetch(threads, id);
+		if (ret)
+			thread_db_incref(threads, id);
+		MUTEX_UNLOCK(&threads_lock);
+		if (!ret)
+			Perl_croak(aTHX_  "Thread %"UVuf" doesn't exist", id);
+	}
+	if (!ret->alive)
+		Perl_croak(aTHX_ "Thread is no longer alive");
+	return ret;
 }
 
 #define get_thread(id) S_get_thread(aTHX_ id)
 
-UV S_queue_alloc(pTHX_ IV linked_to) {
-	message_queue* queue;
-	queue = PerlMemShared_calloc(1, sizeof *queue);
-	queue_init(queue);
-	return resource_addobject(&queues, queue);
+UV S_queue_alloc(pTHX) {
+	message_queue* queue = queue_simple_alloc();
+	UV id = generator();
+	queue_db_store(queues, id, queue);
+	return id;
 }
 
-static message_queue* S_get_queue(pTHX_ UV queue_id) {
-	if (queue_id >= queues.current || queues.objects[queue_id] == NULL)
-		Perl_croak(aTHX_  "queue %"UVuf" doesn't exist", queue_id);
-	return queues.objects[queue_id];
-}
-
-#define get_queue(id) S_get_queue(aTHX_ id)
+#define get_queue(id) queue_db_fetch(queues, id)
 
 #define THREAD_TRY \
 	XCPT_TRY_START
@@ -169,32 +166,29 @@ static message_queue* S_get_queue(pTHX_ UV queue_id) {
 void S_thread_send(pTHX_ UV thread_id, const message* message) {
 	dXCPT;
 
-	MUTEX_LOCK(&threads.lock);
-	THREAD_TRY {
-		mthread* thread = get_thread(thread_id);
-		queue_enqueue(&thread->queue, message, &threads.lock);
-	} THREAD_CATCH( MUTEX_UNLOCK(&threads.lock) );
+	mthread* thread = get_thread(thread_id);
+	queue_enqueue(thread->queue, message, NULL);
 }
 
 void S_queue_send(pTHX_ UV queue_id, const message* message) {
 	dXCPT;
 
-	MUTEX_LOCK(&queues.lock);
+	MUTEX_LOCK(&queues_lock);
 	THREAD_TRY {
 		message_queue* queue = get_queue(queue_id);
-		queue_enqueue(queue, message, &queues.lock);
-	} THREAD_CATCH( MUTEX_UNLOCK(&queues.lock) );
+		queue_enqueue(queue, message, &queues_lock);
+	} THREAD_CATCH( MUTEX_UNLOCK(&queues_lock) );
 }
 
 const message* S_queue_receive(pTHX_ UV queue_id) {
 	dXCPT;
 	const message* ret;
 
-	MUTEX_LOCK(&queues.lock);
+	MUTEX_LOCK(&queues_lock);
 	THREAD_TRY {
 		message_queue* queue = get_queue(queue_id);
-		ret = queue_dequeue(queue, &queues.lock);
-	} THREAD_CATCH( MUTEX_UNLOCK(&queues.lock) );
+		ret = queue_dequeue(queue, &queues_lock);
+	} THREAD_CATCH( MUTEX_UNLOCK(&queues_lock) );
 
 	return ret;
 }
@@ -203,11 +197,11 @@ const message* S_queue_receive_nb(pTHX_ UV queue_id) {
 	dXCPT;
 	const message* ret;
 
-	MUTEX_LOCK(&queues.lock);
+	MUTEX_LOCK(&queues_lock);
 	THREAD_TRY {
 		message_queue* queue = get_queue(queue_id);
-		ret = queue_dequeue_nb(queue, &queues.lock);
-	} THREAD_CATCH( MUTEX_UNLOCK(&queues.lock) );
+		ret = queue_dequeue_nb(queue, &queues_lock);
+	} THREAD_CATCH( MUTEX_UNLOCK(&queues_lock) );
 	return ret;
 }
 
@@ -219,12 +213,15 @@ void S_send_listeners(pTHX_ mthread* thread, const message* mess) {
 	for (i = 0; i < thread->listeners.head; ++i) {
 		const message* clone;
 		UV thread_id;
-		MUTEX_LOCK(&threads.lock); /* unlocked by queue_enqueue */
+		mthread* receptient;
+
+		MUTEX_LOCK(&threads_lock); /* unlocked by queue_enqueue */
 		thread_id = thread->listeners.list[i];
-		if (thread_id >= threads.current || threads.objects[thread_id] == NULL)
+		receptient = thread_db_fetch(threads, thread_id);
+		if (receptient == NULL)
 			continue;
 		clone = message_clone(mess);
-		queue_enqueue(&((mthread*)threads.objects[thread_id])->queue, clone, &threads.lock);
+		queue_enqueue(receptient->queue, clone, &threads_lock);
 	}
 	MUTEX_UNLOCK(&thread->lock);
 }
@@ -234,14 +231,14 @@ void S_send_listeners(pTHX_ mthread* thread, const message* mess) {
 void thread_add_listener(pTHX, UV talker, UV listener) {
 	dXCPT;
 
-	MUTEX_LOCK(&threads.lock);
+	mthread* thread = get_thread(talker);
+	MUTEX_LOCK(&thread->lock);
 	THREAD_TRY {
-		mthread* thread = get_thread(talker);
 		if (thread->listeners.alloc == thread->listeners.head) {
 			thread->listeners.alloc = thread->listeners.alloc ? thread->listeners.alloc * 2 : 1;
 			thread->listeners.list = PerlMemShared_realloc(thread->listeners.list, sizeof(IV) * thread->listeners.alloc);
 		}
 		thread->listeners.list[thread->listeners.head++] = listener;
-	} THREAD_FINALLY( MUTEX_UNLOCK(&threads.lock) );
+	} THREAD_FINALLY( MUTEX_UNLOCK(&thread->lock) );
 }
 
